// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
using System.Collections.Generic;
using Bicep.Core.UnitTests.Assertions;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using FluentAssertions;
using System;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using Bicep.Core.UnitTests.Utils;
using Bicep.Core.FileSystem;
using FluentAssertions.Execution;
using System.Text.RegularExpressions;
using Bicep.Decompiler.Exceptions;
using Bicep.Decompiler;
using Bicep.Core.UnitTests;
using Bicep.Core.UnitTests.Baselines;
using System.Threading;
using System.Globalization;
using Bicep.Core.UnitTests.FileSystem;
using System.Threading.Tasks;

namespace Bicep.Core.IntegrationTests
{
    [TestClass]
    public class DecompilationTests
    {
        [NotNull]
        public TestContext? TestContext { get; set; }

        private static BicepDecompiler CreateDecompiler()
            => ServiceBuilder.Create(s => s.WithEmptyAzResources()).GetDecompiler();

        [DataTestMethod]
        [EmbeddedFilesTestData(@"Files/Working/.*\.json")]
        [TestCategory(BaselineHelper.BaselineTestCategory)]
        public async Task Decompiler_generates_expected_bicep_files_with_diagnostics(EmbeddedFile embeddedJson)
        {
            var baselineFolder = BaselineFolder.BuildOutputFolder(TestContext, embeddedJson);
            var jsonFile = baselineFolder.EntryFile;

            var jsonUri = PathHelper.FilePathToFileUrl(jsonFile.OutputFilePath);
            var decompiler = ServiceBuilder.Create().GetDecompiler();
            var (bicepUri, filesToSave) = await decompiler.Decompile(PathHelper.ChangeToBicepExtension(jsonUri), jsonFile.EmbeddedFile.Contents);

            var result = CompilationHelper.Compile(new ServiceBuilder().BuildCompilation(filesToSave, PathHelper.ChangeToBicepExtension(jsonUri)));
            var diagnosticsByBicepFile = result.Compilation.GetAllDiagnosticsByBicepFile();

            using (new AssertionScope())
            {
                foreach (var (bicepFile, diagnostics) in diagnosticsByBicepFile)
                {
                    var baselineFile = baselineFolder.GetFileOrEnsureCheckedIn(bicepFile.FileUri);
                    var bicepOutput = filesToSave[bicepFile.FileUri];

                    var sourceTextWithDiags = OutputHelper.AddDiagsToSourceText(bicepOutput, "\n", diagnostics, diag => OutputHelper.GetDiagLoggingString(bicepOutput, baselineFolder.OutputFolderPath, diag));

                    baselineFile.WriteToOutputFolder(sourceTextWithDiags);
                    baselineFile.ShouldHaveExpectedValue();
                }
            }
        }

        [DataTestMethod]
        [EmbeddedFilesTestData(@"Files/Parameters/.*\.json")]
        [TestCategory(BaselineHelper.BaselineTestCategory)]
        public void Decompiler_generates_expected_bicepparam_files_with_diagnostics(EmbeddedFile embeddedJson)
        {
            var baselineFolder = BaselineFolder.BuildOutputFolder(TestContext, embeddedJson);
            var jsonFile = baselineFolder.EntryFile;

            var jsonUri = PathHelper.FilePathToFileUrl(jsonFile.OutputFilePath);
            var decompiler = ServiceBuilder.Create().GetBicepparamDecompiler();

            var (entryPointUri, filesToSave) = decompiler.Decompile(jsonUri, PathHelper.ChangeExtension(jsonUri, LanguageConstants.ParamsFileExtension), null);

            var baselineFile = baselineFolder.GetFileOrEnsureCheckedIn(entryPointUri);
            baselineFile.WriteToOutputFolder(filesToSave[entryPointUri]);
            baselineFile.ShouldHaveExpectedValue();
        }

        private static string ReadResourceFile(string resourcePath)
        {
            var manifestStream = typeof(DecompilationTests).Assembly.GetManifestResourceStream(resourcePath)!;
            return new StreamReader(manifestStream).ReadToEnd();
        }

        [DataTestMethod]
        [DataRow("Files/NonWorking/unknownprops.json", "[15:29]: Unrecognized top-level resource property 'madeUpProperty'")]
        [DataRow("Files/NonWorking/invalid-schema.json", "[2:98]: $schema value \"https://schema.management.azure.com/schemas/2019-04-01/deploymentParameters.json#\" did not match any of the known ARM template deployment schemas.")]
        [DataRow("Files/NonWorking/keyvault-secret-reference.json", "[25:38]: Failed to convert parameter \"mySecret\": KeyVault secret references are not currently supported by the decompiler.")]
        [DataRow("Files/NonWorking/symbolic-names.json", "[19:16]: Decompilation of symbolic name templates is not currently supported")]
        public async Task Decompiler_raises_errors_for_unsupported_features(string resourcePath, string expectedMessage)
        {
            Func<Task> onDecompile = async () =>
            {
                var jsonContent = ReadResourceFile(resourcePath);
                var decompiler = CreateDecompiler();
                await decompiler.Decompile(new Uri("file:///unused.bicep"), jsonContent);
            };

            await onDecompile.Should().ThrowAsync<ConversionFailedException>().WithMessage(expectedMessage);
        }

        [DataTestMethod]
        [DataRow("\r\n", "\\r\\n")]
        [DataRow("\n", "\\n")]
        public async Task Decompiler_handles_strings_with_newlines(string newline, string escapedNewline)
        {
            var template = @"{
    ""$schema"": ""https://schema.management.azure.com/schemas/2019-04-01/deploymentTemplate.json#"",
    ""contentVersion"": ""1.0.0.0"",
    ""parameters"": {},
    ""variables"": {
        ""multilineString"": ""multi
        line
        string""
    },
    ""resources"": [],
    ""outputs"": {}
}";

            // replace newlines with the style passed in
            template = string.Join(newline, Regex.Split(template, "\r?\n"));

            var fileUri = new Uri("file:///path/to/main.json");

            var decompiler = CreateDecompiler();
            var (entryPointUri, filesToSave) = await decompiler.Decompile(PathHelper.ChangeToBicepExtension(fileUri), template);

            // this behavior is actually controlled by newtonsoft's deserializer, but we should assert it anyway to avoid regressions.
            filesToSave[entryPointUri].Should().Contain($"var multilineString = 'multi{escapedNewline}        line{escapedNewline}        string'");
        }

        [DataTestMethod]
        [DataRow("and(variables('a'), variables('b'))", "boolean", "(a && b)")]
        [DataRow("and(variables('a'), variables('b'), variables('c'))", "boolean", "(a && b && c)")]
        [DataRow("or(variables('a'), variables('b'))", "boolean", "(a || b)")]
        [DataRow("or(variables('a'), variables('b'), variables('c'))", "boolean", "(a || b || c)")]
        [DataRow("add(variables('a'), variables('b'))", "int", "(a + b)")]
        [DataRow("sub(variables('a'), variables('b'))", "int", "(a - b)")]
        [DataRow("mul(variables('a'), variables('b'))", "int", "(a * b)")]
        [DataRow("div(variables('a'), variables('b'))", "int", "(a / b)")]
        [DataRow("mod(variables('a'), variables('b'))", "int", "(a % b)")]
        [DataRow("less(variables('a'), variables('b'))", "boolean", "(a < b)")]
        [DataRow("lessOrEquals(variables('a'), variables('b'))", "boolean", "(a <= b)")]
        [DataRow("greater(variables('a'), variables('b'))", "boolean", "(a > b)")]
        [DataRow("greaterOrEquals(variables('a'), variables('b'))", "boolean", "(a >= b)")]
        [DataRow("equals(variables('a'), variables('b'))", "boolean", "(a == b)")]
        [DataRow("equals(toLower(variables('a')),toLower(variables('b')))", "boolean", "(a =~ b)")]
        [DataRow("not(equals(variables('a'),variables('b')))", "boolean", "(a != b)")]
        [DataRow("not(equals(toLower(variables('a')),toLower(variables('b'))))", "boolean", "(a !~ b)")]
        [DataRow("createArray(1, 2, 3)", "array", "[\n  1\n  2\n  3\n]")]
        [DataRow("createObject('key', 'value')", "object", "{\n  key: 'value'\n}")]
        [DataRow("tryGet(parameters('z'), 'y')", "int", "(z.?y)")]
        [DataRow("tryGet(parameters('z'), 'y', 'x', 'w')", "int", "(z.?y.x.w)")]
        [DataRow("tryGet(tryGet(parameters('z'), 'y', 'x'), 'w', 'v')", "int", "((z.?y.x).?w.v)")]
        [DataRow("tryGet(parameters('z'), 'y', 'x', 'w').v", "int", "(z.?y.x.w).v")]
        public async Task Decompiler_handles_banned_function_replacement(string expression, string type, string expectedValue)
        {
            var template = @"{
    ""$schema"": ""https://schema.management.azure.com/schemas/2019-04-01/deploymentTemplate.json#"",
    ""contentVersion"": ""1.0.0.0"",
    ""parameters"": {
        ""z"": {""type"":""object""}
    },
    ""variables"": {
        ""a"": true,
        ""b"": false,
        ""c"": true
    },
    ""resources"": [],
    ""outputs"": {
        ""calculated"": {
            ""type"": """ + type + @""",
            ""value"": ""[" + expression + @"]""
        }
    }
}";

            var fileUri = new Uri("file:///path/to/main.json");

            var decompiler = CreateDecompiler();
            var (entryPointUri, filesToSave) = await decompiler.Decompile(PathHelper.ChangeToBicepExtension(fileUri), template);

            filesToSave[entryPointUri].Should().Contain($"output calculated {type} = {expectedValue}");
        }

        [TestMethod]
        public async Task Decompiler_should_partially_handle_user_defined_functions_with_placeholders()
        {
            const string template = @"{
 ""$schema"": ""https://schema.management.azure.com/schemas/2019-04-01/deploymentTemplate.json#"",
 ""contentVersion"": ""1.0.0.0"",
 ""parameters"": {
   ""storageNamePrefix"": {
     ""type"": ""string"",
     ""maxLength"": 11
   }
 },
 ""functions"": [
  {
    ""namespace"": ""contoso"",
    ""members"": {
      ""uniqueName"": {
        ""parameters"": [
          {
            ""name"": ""namePrefix"",
            ""type"": ""string""
          }
        ],
        ""output"": {
          ""type"": ""string"",
          ""value"": ""[concat(toLower(parameters('namePrefix')), uniqueString(resourceGroup().id))]""
        }
      }
    }
  }
],
 ""resources"": [
   {
     ""type"": ""Microsoft.Storage/storageAccounts"",
     ""apiVersion"": ""2019-04-01"",
     ""name"": ""[contoso.uniqueName(parameters('storageNamePrefix'))]"",
     ""location"": ""South Central US"",
     ""sku"": {
       ""name"": ""Standard_LRS""
     },
     ""kind"": ""StorageV2"",
     ""properties"": {
       ""supportsHttpsTrafficOnly"": true
     }
   }
 ]
}";

            var fileUri = new Uri("file:///path/to/main.json");

            var decompiler = CreateDecompiler();
            var (entryPointUri, filesToSave) = await decompiler.Decompile(PathHelper.ChangeToBicepExtension(fileUri), template);

            filesToSave[entryPointUri].Should().Contain($"? /* TODO: User defined functions are not supported and have not been decompiled */");
        }

        [TestMethod]
        public async Task Decompiler_should_not_interpret_numbers_with_locale_settings()
        {
            // https://github.com/Azure/bicep/issues/7615
            const string template = @"{
    ""$schema"": ""https://schema.management.azure.com/schemas/2019-04-01/deploymentTemplate.json#"",
    ""contentVersion"": ""1.0.0.0"",
    ""parameters"": {},
    ""variables"": {
        ""cpu"": 0.25
    },
    ""resources"": [],
    ""outputs"": {}
}";

            var fileUri = new Uri("file:///path/to/main.json");

            var currentCulture = Thread.CurrentThread.CurrentCulture;
            try {
                Thread.CurrentThread.CurrentCulture = new CultureInfo("fi-FI");

                var decompiler = CreateDecompiler();
                var (entryPointUri, filesToSave) = await decompiler.Decompile(PathHelper.ChangeToBicepExtension(fileUri), template);

                filesToSave[entryPointUri].Should().Contain($"var cpu = '0.25'");
            }
            finally {
                Thread.CurrentThread.CurrentCulture = currentCulture;
            }
        }
    }
}
