// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
using System;
using System.IO;
using System.Collections.Generic;
using Microsoft.JSInterop;
using Bicep.Core.Diagnostics;
using Bicep.Core.Text;
using Bicep.Core.Emit;
using Bicep.Core.Semantics;
using Bicep.Wasm.LanguageHelpers;
using System.Linq;
using Bicep.Core.FileSystem;
using Bicep.Core.Extensions;
using Bicep.Decompiler;
using Microsoft.Extensions.DependencyInjection;
using System.IO.Abstractions;
using System.Threading.Tasks;
using Bicep.Core;

namespace Bicep.Wasm
{
    public class Interop
    {
        public record DecompileResult(string? bicepFile, string? error);

        public record CompileResult(string template, object diagnostics);

        private readonly IJSRuntime jsRuntime;
        private readonly IServiceProvider serviceProvider;

        public Interop(IJSRuntime jsRuntime, IServiceProvider serviceProvider)
        {
            this.jsRuntime = jsRuntime;
            this.serviceProvider = serviceProvider;
        }

        [JSInvokable]
        public async Task<CompileResult> CompileAndEmitDiagnostics(string content)
        {
            try
            {
                var compilation = await GetCompilation(content);
                var lineStarts = compilation.SourceFileGrouping.EntryPoint.LineStarts;
                var emitter = new TemplateEmitter(compilation.GetEntrypointSemanticModel());

                var stringWriter = new StringWriter();
                var emitResult = emitter.Emit(stringWriter);

                if (emitResult.Status != EmitStatus.Failed)
                {
                    // compilation was successful or had warnings - return the compiled template
                    return new(stringWriter.ToString(), emitResult.Diagnostics.Select(d => ToMonacoDiagnostic(d, lineStarts)));
                }

                // compilation failed
                return new("Compilation failed!", emitResult.Diagnostics.Select(d => ToMonacoDiagnostic(d, lineStarts)));
            }
            catch (Exception exception)
            {
                return new(exception.ToString(), Enumerable.Empty<object>());
            }
        }

        [JSInvokable]
        public async Task<DecompileResult> Decompile(string jsonContent)
        {
            using var serviceScope = serviceProvider.CreateScope();
            var decompiler = serviceScope.ServiceProvider.GetRequiredService<BicepDecompiler>();

            try
            {
                var (entrypointUri, filesToSave) = await decompiler.Decompile(new Uri("inmemory:///main.bicep"), jsonContent);

                return new DecompileResult(filesToSave[entrypointUri], null);
            }
            catch (Exception exception)
            {
                return new DecompileResult(null, exception.Message);
            }
        }

        [JSInvokable]
        public object GetSemanticTokensLegend()
        {
            var tokenTypes = Enum.GetValues(typeof(SemanticTokenType)).Cast<SemanticTokenType>();
            var tokenStrings = tokenTypes.OrderBy(t => (int)t).Select(t => t.ToString().ToLowerInvariant());

            return new
            {
                tokenModifiers = new string[] { },
                tokenTypes = tokenStrings.ToArray(),
            };
        }

        [JSInvokable]
        public async Task<object> GetSemanticTokens(string content)
        {
            var compilation = await GetCompilation(content);
            var tokens = SemanticTokenVisitor.BuildSemanticTokens(compilation.GetEntrypointSemanticModel());

            var data = new List<int>();
            SemanticToken? prevToken = null;
            foreach (var token in tokens)
            {
                if (prevToken == null)
                {
                    data.Add(token.Line);
                    data.Add(token.Character);
                    data.Add(token.Length);
                }
                else if (prevToken.Line != token.Line)
                {
                    data.Add(token.Line - prevToken.Line);
                    data.Add(token.Character);
                    data.Add(token.Length);
                }
                else
                {
                    data.Add(0);
                    data.Add(token.Character - prevToken.Character);
                    data.Add(token.Length);
                }

                data.Add((int)token.TokenType);
                data.Add(0);

                prevToken = token;
            }

            return new
            {
                data = data.ToArray(),
            };
        }

        private async Task<Compilation> GetCompilation(string fileContents)
        {
            using var serviceScope = serviceProvider.CreateScope();
            var compiler = serviceScope.ServiceProvider.GetRequiredService<BicepCompiler>();
            var fileSystem = serviceScope.ServiceProvider.GetRequiredService<IFileSystem>();

            var fileUri = new Uri("file:///main.bicep");
            await fileSystem.File.WriteAllTextAsync(fileUri.LocalPath, fileContents);

            return await compiler.CreateCompilation(fileUri);
        }

        private static object ToMonacoDiagnostic(IDiagnostic diagnostic, IReadOnlyList<int> lineStarts)
        {
            var (startLine, startChar) = TextCoordinateConverter.GetPosition(lineStarts, diagnostic.Span.Position);
            var (endLine, endChar) = TextCoordinateConverter.GetPosition(lineStarts, diagnostic.GetEndPosition());

            return new
            {
                code = diagnostic.Code,
                message = diagnostic.Message,
                severity = ToMonacoSeverity(diagnostic.Level),
                startLineNumber = startLine + 1,
                startColumn = startChar + 1,
                endLineNumber = endLine + 1,
                endColumn = endChar + 1,
            };
        }

        private static int ToMonacoSeverity(DiagnosticLevel level)
            => level switch
            {
                DiagnosticLevel.Info => 2,
                DiagnosticLevel.Warning => 4,
                DiagnosticLevel.Error => 8,
                _ => throw new ArgumentException($"Unrecognized level {level}"),
            };
    }
}
