// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
import vscode from "vscode";
import path from "path";
import crypto from "crypto";
import { LanguageClient } from "vscode-languageclient/node";

import { createDeploymentGraphMessage, Message } from "./messages";
import { deploymentGraphRequestType } from "../language";
import { parseError } from "@microsoft/vscode-azext-utils";
import { Disposable } from "../utils/disposable";
import { debounce } from "../utils/time";
import { getLogger } from "../utils/logger";

export class BicepVisualizerView extends Disposable {
  public static viewType = "bicep.visualizer";

  private readonly onDidDisposeEmitter: vscode.EventEmitter<void>;
  private readonly onDidChangeViewStateEmitter: vscode.EventEmitter<vscode.WebviewPanelOnDidChangeViewStateEvent>;

  private readyToRender = false;

  private constructor(
    private readonly languageClient: LanguageClient,
    private readonly webviewPanel: vscode.WebviewPanel,
    private readonly extensionUri: vscode.Uri,
    private readonly documentUri: vscode.Uri,
  ) {
    super();

    this.onDidDisposeEmitter = new vscode.EventEmitter<void>();
    this.onDidChangeViewStateEmitter = this.register(
      new vscode.EventEmitter<vscode.WebviewPanelOnDidChangeViewStateEvent>(),
    );

    this.register(
      this.webviewPanel.webview.onDidReceiveMessage(
        // eslint-disable-next-line jest/unbound-method
        this.handleDidReceiveMessage,
        this,
      ),
    );

    if (!this.isDisposed) {
      this.webviewPanel.webview.html = this.createWebviewHtml();
    }

    this.registerMultiple(
      // eslint-disable-next-line jest/unbound-method
      this.webviewPanel.onDidDispose(this.dispose, this),
      this.webviewPanel.onDidChangeViewState((e) =>
        this.onDidChangeViewStateEmitter.fire(e),
      ),
    );
  }

  public get onDidDispose(): vscode.Event<void> {
    return this.onDidDisposeEmitter.event;
  }

  public get onDidChangeViewState(): vscode.Event<vscode.WebviewPanelOnDidChangeViewStateEvent> {
    return this.onDidChangeViewStateEmitter.event;
  }

  public static create(
    languageClient: LanguageClient,
    viewColumn: vscode.ViewColumn,
    extensionUri: vscode.Uri,
    documentUri: vscode.Uri,
  ): BicepVisualizerView {
    const visualizerTitle = `Visualize ${path.basename(documentUri.fsPath)}`;
    const webviewPanel = vscode.window.createWebviewPanel(
      BicepVisualizerView.viewType,
      visualizerTitle,
      viewColumn,
      {
        enableScripts: true,
        retainContextWhenHidden: true,
      },
    );

    return new BicepVisualizerView(
      languageClient,
      webviewPanel,
      extensionUri,
      documentUri,
    );
  }

  public static revive(
    languageClient: LanguageClient,
    webviewPanel: vscode.WebviewPanel,
    extensionUri: vscode.Uri,
    documentUri: vscode.Uri,
  ): BicepVisualizerView {
    return new BicepVisualizerView(
      languageClient,
      webviewPanel,
      extensionUri,
      documentUri,
    );
  }

  public reveal(): void {
    this.webviewPanel.reveal();
  }

  public dispose(): void {
    super.dispose();

    this.webviewPanel.dispose();

    // Final cleanup.
    this.onDidDisposeEmitter.fire();
    this.onDidDisposeEmitter.dispose();
  }

  // Do "fire and forget" since there's no need to wait on rendering.
  public render = debounce(() => this.doRender());

  private async doRender() {
    if (this.isDisposed || !this.readyToRender) {
      return;
    }

    let document: vscode.TextDocument;
    try {
      document = await vscode.workspace.openTextDocument(this.documentUri);
    } catch {
      this.webviewPanel.webview.html = this.createDocumentNotFoundHtml();
      return;
    }

    if (this.isDisposed) {
      return;
    }

    const deploymentGraph = await this.languageClient.sendRequest(
      deploymentGraphRequestType,
      {
        textDocument:
          this.languageClient.code2ProtocolConverter.asTextDocumentIdentifier(
            document,
          ),
      },
    );

    if (this.isDisposed) {
      return;
    }

    try {
      await this.webviewPanel.webview.postMessage(
        createDeploymentGraphMessage(this.documentUri.fsPath, deploymentGraph),
      );
    } catch (error) {
      // Race condition: the webview was closed before receiving the message,
      // which causes "Unknown webview handle" error.
      getLogger().debug((error as Error).message ?? error);
    }
  }

  private handleDidReceiveMessage(message: Message): void {
    switch (message.kind) {
      case "READY":
        getLogger().debug(
          `Visualizer for ${this.documentUri.fsPath} is ready.`,
        );

        this.readyToRender = true;
        this.render();

        return;
      case "REVEAL_FILE_RANGE":
        this.revealFileRange(message.filePath, message.range);

        return;
    }
  }

  private revealFileRange(filePath: string, range: vscode.Range) {
    for (const visibleEditor of vscode.window.visibleTextEditors) {
      if (visibleEditor.document.uri.fsPath === filePath) {
        vscode.window
          .showTextDocument(visibleEditor.document, visibleEditor.viewColumn)
          .then(
            (editor) => this.revealEditorRange(editor, range),
            (err) =>
              vscode.window.showErrorMessage(
                `Could not reveal file range in "${filePath}": ${
                  parseError(err).message
                }`,
              ),
          );
        return;
      }
    }

    vscode.workspace
      .openTextDocument(filePath)
      .then(vscode.window.showTextDocument)
      .then(
        (editor) => this.revealEditorRange(editor, range),
        (err) =>
          vscode.window.showErrorMessage(
            `Could not open "${filePath}": ${parseError(err).message}`,
          ),
      );
  }

  private revealEditorRange(editor: vscode.TextEditor, range: vscode.Range) {
    // editor.selection.active is the current cursor position which is immutable.
    const cursorPosition = editor.selection.active.with(
      range.start.line,
      range.start.character,
    );
    // Move cursor to the beginning of the resource/module and reveal the source code.
    editor.selection = new vscode.Selection(cursorPosition, cursorPosition);
    editor.revealRange(range, vscode.TextEditorRevealType.InCenter);
  }

  private createWebviewHtml() {
    const { cspSource } = this.webviewPanel.webview;
    const nonce = crypto.randomBytes(16).toString("hex");
    const scriptUri = this.webviewPanel.webview.asWebviewUri(
      vscode.Uri.joinPath(this.extensionUri, "out", "visualizer.js"),
    );

    return `
      <!DOCTYPE html>
      <html lang="en">
      <head>
        <meta charset="UTF-8">
        <!--
        Use a content security policy to only allow loading images from our extension directory,
        and only allow scripts that have a specific nonce.
        -->
        <meta http-equiv="Content-Security-Policy" content="default-src 'none'; style-src ${cspSource} 'unsafe-inline'; img-src ${cspSource} data:; script-src 'nonce-${nonce}' vscode-webview-resource:;">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
      </head>
      <body>
        <div id="root"></div>
        <script nonce="${nonce}" src="${scriptUri}" />
      </body>
      </html>`;
  }

  private createDocumentNotFoundHtml() {
    const documentName = path.basename(this.documentUri.fsPath);

    return `
      <!DOCTYPE html>
      <html lang="en">
      <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
      </head>
      <body>
        <div class="vscode-body">${documentName} not found. It might be deleted or renamed.</div>
      </body>
      </html>`;
  }
}
